#! /usr/bin/perl
#
# Copyright (c) 2013, Richard Appleby
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * The name of Richard Appleby may not be used to endorse or promote products
#      derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL RICHARD APPLEBY BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Function:
#
# Retrieves and logs the weather stats from a Fine Offset WH-1080/1081 weather
# station using the USB connector on the base station, prints the results, and
# optionally uploads them to a Weather Underground PWS.
#
# In the UK these weatherstations are marketed via Maplin as product code N96FY
#
# Forked and extensively modifed (with permission) from an original concept 
# by Steve Cliffe located at http://www.sjcnet.id.au/computers/weather-station-data-logger
# 
# 1.0 - 22 Nov 2013, First released at http://theapplebyclan.com/Richard/Weatherstation/
# 1.1 - 06 Dec 2014, Added better error checking to WeatherUnderground upload routine

# Needed for upload to WeatherUnderground
use LWP::UserAgent;
use HTTP::Request;
# Needed for general function
use POSIX qw(sysconf pause strftime);
use Device::USB;
use Data::Dumper;
use Carp::Assert;
use Getopt::Std;
use Time::HiRes qw(setitimer ITIMER_REAL);

# USB identifiers for the Fine Offset weather stations, allowing us to connect to the right
# device
my $VENDOR = 0x1941;
my $PRODUCT = 0x8021;

# Set to 1 to have additional information written to the output
my $debug = 0;	

# Try to improve the madly swinging wind direction measurements by keeping a running average
# of some number of historical readings. Given the inaccuracies in the basic wind vane, this
# may well give you a (subjectively) better result. Set to 1 to enable, set to 0 to use the raw data.
my $enable_wind_avg = 1;

#
# WeatherUnderground upload configuration:
#

# If you want to upload your results to WeatherUnderground, set this to 1 ...
my $wuUpload = 1;
# ... and add your WeatherUnderground PWS configuration information here:
my $ID = "";
my $PASSWORD = "";
my $WU_URL = "http://weatherstation.wunderground.com/weatherstation/updateweatherstation.php";
my $SOFTWARETYPE = "Custom PERL code";

#
# Command line arguments
#
usage() if defined $ARGV[0] and $ARGV[0] eq "--help";
getopts('dhv') or usage();
usage() if defined $main::opt_h;
revision() if defined $main::opt_v;
my $debug = defined $main::opt_d ? $main::opt_d : 0;

# process [[interval [count]]
my ($interval, $loop_max);
if (defined $ARGV[0]) {
    $interval = $ARGV[0];
    $loop_max = defined $ARGV[1] ? $ARGV[1] : 2**32;
    usage() if $interval == 0;
    }
else {
    $interval = 1;
    $loop_max = 1; 
    }

my $loop = 0;           # current loop number
my $PAGESIZE = 25;      # max lines of output between headers
my $line = $PAGESIZE;   # counter for lines printed

$main::opt_h = 0;
$main::opt_v = 0;
$main::opt_d = 0;
$| = 1;                    # autoflush


##
## usage - print usage and exit.
##
sub usage {
        print STDERR <<END;
USAGE: 
	wsreader [-v] | [-h | --help] | [-d] [interval [count]]

OPTIONS:
	-v		Print version information
	-h | --help	Print (this) help
	-d		Print extra debugging information

EXAMPLES:
	wsreader      	Print a reading and stop
	wsreader 5	Print a reading every 5 seconds
	wsreader 1 5	Print a reading every second, stop after 5 readings

OUTPUT FIELDS:
	Time    	Current time
	%IH     	% Indoor humidity
	%OH     	% Outdoor humidity
	IT°C    	Indoor temp Celsius 
	OT°C    	Outdoor temp Celsius
	DP°C    	Outdoor dew point temp Celsius
	WC°C    	Outdoor wind chill temp Celsius
	Wm/s    	Wind speed meter/s
	Gust    	Wind gust meter/s
	WD      	Wind direction (8 ordinals, converted to degrees)
	1hr     	Rainfall in the last hour mm
	24hr    	Rainfall today (since local midnight) mm
	T       	Rain total (raw data from console) mm
	P       	Abs Pressure hPa
END
  exit 1;
}

##
## revision - print revision and exit
##
sub revision {
       print STDERR <<END;
wsreader: 1.1, 2013-12-06
END
       exit 1;
}


##
## Connect to the Weather Station
##
sub open_ws {

	my $usb = shift;
	my $namebuf = "\0" x 256;
	my $dev = $usb->find_device($VENDOR, $PRODUCT);

	if ($dev->get_driver_np(0, $namebuf, 256) == 0) {
        	$dev->detach_kernel_driver_np(0);
	    	}	

	$dev->open();

	if ($dev->claim_interface(0) != 0) {
		printf "usb_claim_interface failed\n";
		return 0;
		}
	$dev->set_altinterface(0);

	return $dev;
}

##
## Close the connection to the Weather Station & exit
##
sub close_ws {
	$dev->release_interface(0);
	undef $dev;
	exit;
}

##
## Dump a buffer in hex for debugging purposes
##
sub print_bytes {
	my $buf = shift;
	my $len = shift;
   
	if ($len <= 0) {
		return;
	}
	my @bytes = unpack("C$len", $buf);

    	if ($len > 0) {
        	for (my $i=0; $i<$len; $i++) {
            		printf "%02x ", $bytes[$i];
	        	}
	    	}
	printf "\n";
}

##
## Read a 32 byte block of data from the WS
## (Cannot reliably read any other size)
##
sub read_block {
	my $dev = shift;
	my $offset = shift;

        my $xbuf = "\0" x 256;

	my $readCommand = 0xA1;
	my $endMark = 0x20;
	my $lsb = $offset & 0xFF;
	my $msb = $offset >>8 & 0xFF;
	# Pack the read command into the buffer
	my $tbuf = pack('CCCCCCCC', $readCommand, $msb, $lsb, $endMark, $readCommand, $msb, $lsb, $endMark);
	# Send command, using a USB control_msg to the weatherstation:
	# --- control_msg(requesttype, request, value, index, buffer, sizeof(buffer), timeout)
	# --- requesttype = ENDPOINT_OUT | CTRL_TYPE_CLASS | CTRL_RECIPIENT_INTERFACE
	# --- request = REQ_SET_CONFIGURATION
	my $retval = $dev->control_msg(0x21, 9, 0x200, 0, $tbuf, 8, 1000);
	# Now read the requested data back from the weatherstation with:
	# --- interrupt_read(endpoint, buffer, length, timeout)
        my $count = $dev->interrupt_read(0x81, $buf, 32, 1000);
	if ($debug) {
                printf ("Retval: %d, Read %d bytes\n", $retval, $count);
	        }

	return $buf;
}


##
## Calculate dew point based on temperature & humidity
##
## http://en.wikipedia.org/wiki/Dew_Point
##
sub dew_point {
	my $temp = shift;
	my $humidity = shift;
	$humidity /= 100.0;
	my $gamma = (17.271 * $temp) / (237.7 + $temp) + log($humidity) ;
	return (237.7 * $gamma) / (17.271 - $gamma);
}

##
## Calculate wind chill temp based on temperature & wind speed
##
## http://en.wikipedia.org/wiki/Wind_chill
##
sub wind_chill {
	my $temp = shift;
	my $wind = shift;

	my $wind_kph = 3.6 * $wind;
	if (($wind_kph <= 4.8) || ($temp > 10.0)) {
		return $temp;
		}
	my $wct = 13.12 + (0.6215 * $temp) - (11.37 * ($wind_kph ** 0.16)) + (0.3965 * $temp * ($wind_kph ** 0.16));
	if ($wct < $temp) {
		return $wct;
		} 
	else {
		return $temp;
		}
}

##
## Conversion routines needed to switch from metric units in the Fine Offset
## weather stations to the imperial units used by WeatherUnderground
##

# mstomph : Convert from meters per second to miles per houra (Windspeed)
sub mstomph ($) {
  my ($ms) = @_;
  return sprintf("%.2f",$ms*2.23693629);
}

# CtoF : Convert from degrees Centigrade to degrees Fahrenheit (Temps)
sub CtoF ($) {
  my ($C) = @_;
  return sprintf("%.2f",9*$C/5+32);
}

# mmtoin : Convert from Millimeters to Inches (Rain)
sub mmtoin ($) {
  my ($mm) = @_;
  return sprintf("%.3f",$mm/25.4);
}

# hPatoin : Convert from hectaPascals to InchesHg/Mercury (Pressure)
sub hPatoin ($) {
  my ($hPa) = @_;
  return sprintf("%.3f",$hPa/33.8638864);
}

##
##
## Main Loop
##
##
my @wind_dirs = ('N', 'NNE', 'NE', 'ENE', 'E', 'ESE', 'SE', 'SSE', 'S', 'SSW', 'SW', 'WSW', 'W', 'WNW', 'NW', 'NNW');
my @wind_dirs_degs = ('0', '22', '45', '67', '90', '112', '135', '157', '180', '202', '225', '247', '270', '292', '315', '337');
my @wind_history;	# Array of historical wind directions, in degrees, for averaging, if required
my $hourly_rain_rate = 0;
my $daily_rain_rate = 0;

my $upload_counter, $upload_frequency;

#
# WeatherUnderground wont accept updates more often than every 5 minutes, but
# we may want to run more frequently, so ensure we dont upload too often.
#
if ($interval < 300) {
	if ((300/$interval) == (int(300/$interval))) {
		$upload_frequency = 300 / $interval;
		}
	else {
		$upload_frequency = int(300 / $interval)+1;
		}
	}
else {
	$upload_frequency = 0;
	}

# Trigger immediate upload
$upload_counter = 0 ; 

# how often do we trigger
my $first_interval = $interval;

# signal handler is empty, do nothing
$SIG{ALRM} = sub {};

# first value is the initial wait, second is the wait thereafter
setitimer(ITIMER_REAL, $first_interval, $interval);

$SIG{'INT'} = 'close_ws';
$SIG{'TERM'} = 'close_ws';

#
# Attempt to open a connection to the weather station
#
my $usb = Device::USB->new();
our $dev = open_ws($usb);
if ($dev ==  0) {
	printf "Couldn't connect to the Weather Station\n";
	exit 1;
	}

#
# Loop forever polling the Weather Station
#
while (1) {
	#
	# Get the localtime
	#
	my @Time = localtime();

	#
	# Work out how many packets of saved data the console has recorded since
	# the start of today (midnight), assuming a 5 minute collection rate
	#
	my $PktsToDayStart = int((($Time[2] * 60) + $Time[1]) / 5);

	#
	# Get the first 32 bytes of the 256 byte fixed block at start of memory
	#
	my $fixed_block = read_block($dev, 0);
        if ($debug) {
                print_bytes($fixed_block, 32);
        	}

        if (unpack('C', $fixed_block) != 0x55) {
                printf("Got a dud fixed block - reconnecting\n");
                $dev->release_interface(0);
		undef $dev;
                sleep 10;
                $dev = open_ws($usb);
                next;
        	}

	#
	# Since we depend on the logging interval being set to 5 minutes on the console,
	# which is NOT the default (its normally 30 minutes) we need to check it, and if
	# necessary stop with an informative message on how to fix the problem
	#
	if (unpack('x16C', $fixed_block) != 5) {
		printf("Logging interval on base station needs to be set to 5 minutes!\n");
		printf("See the SetLogger.zip on http://sandaysoft.com/downloads, or\n");
		printf("refer to the pywws documentation, which can be found at\n");
		printf("http://jim-easterbrook.github.io/pywws/doc/en/html/guides/getstarted.html\n");
		exit;
		}

	#
        # Get the 16 byte current weather record
	# This is still (probably) being updated
	#
        my $curpos = unpack('x30S',$fixed_block);
        my $current_block = read_block($dev, $curpos);
        if ($debug) {
		printf ("Current weather record\n");
                print_bytes($current_block, 16);
                printf ("Current position is %x\n", $curpos);
        	}

	#
        # Get the 16 bytes of the previous weather record
	# This is the last complete record, at most 5 minutes old
	#
        my $curpos = $curpos - 16;
	if ($curpos < 256) {
                # Keep us in the right range for weather records
		$curpos = $curpos + 65280;
		}
        my $last_block = read_block($dev, $curpos);
        if ($debug) {
		printf ("Last weather record\n");
                print_bytes($last_block, 16);
                printf ("Current position is %x\n", $curpos);
        	}

	#
        # Get the 16 bytes of the weather record from an hour ago
	# 5 minute recording interval, so 12 blocks back from $last_block
	#
        my $curpos = $curpos - (16 * 12);
	if ($curpos < 256) {
                # Keep us in the right range for weather records
		$curpos = $curpos + 65280;
		}
        my $hour_block = read_block($dev, $curpos);
        if ($debug) {
		printf ("1 hour old weather record\n");
                print_bytes($hour_block, 16);
                printf ("Current position is %x\n", $curpos);
        	}

	#
        # Get the 16 bytes of the weather record from start of (local) day
	#
	my $curpos = unpack('x30S',$fixed_block);
        my $curpos = $curpos - (16 * $PktsToDayStart);
	if ($curpos < 256) {
                # Keep us in the right range for weather records
		$curpos = $curpos + 65280;
		}
        my $day_block = read_block($dev, $curpos);
        if ($debug) {
		printf ("Weather record from start of day\n");
                print_bytes($day_block, 16);
                printf ("Current position is %x\n", $curpos);
        	}

	##
	## Decode humidy, temp, pressure & windspeed/direction from current "inflight" record
	##

	# Indoor humidity and temperature
	my $indoor_humidity = unpack('xC', $current_block);
	my $tlsb = unpack('xxC', $current_block);
 	my $tmsb = unpack('xxxC', $current_block) & 0x7f;
	my $tsign = unpack('xxxC', $current_block) >> 7;
	my $indoor_temperature = ($tmsb * 256 + $tlsb) * 0.1;		# 0.1 for degC
	if ($tsign) {
		$indoor_temperature *= -1;
	}
	
	# Outdoor humidity and temperature
	my $outdoor_humidity = unpack('x4C', $current_block);
        $tlsb = unpack('x5C', $current_block);
        $tmsb = unpack('x6C', $current_block) & 0x7f;
        $tsign = unpack('x6C', $current_block) >> 7;
        $outdoor_temperature = ($tmsb * 256 + $tlsb) * 0.1;		# 0.1 for degC
        if ($tsign) {
                $outdoor_temperature *= -1;
        }

	# Pressure
	my $abs_pressure= unpack('x7S', $current_block) * 0.1;		# 0.1 for hpa

	# Wind/Gust speeds
	my $wind = unpack('x9C', $current_block);
	my $gust = unpack('x10C', $current_block);
	my $wind_extra = unpack('x11C', $current_block);
	$wind_speed = ($wind + (($wind_extra & 0x0F) << 8)) * 0.1;	# 0.1 for m/s
	$gust_speed = ($gust + (($wind_extra & 0xF0) << 4)) * 0.1;	# 0.1 for m/s

	# Wind direction
	my $wind_dir = unpack('x12C', $current_block);
	$wind_dir = @wind_dirs_degs[$wind_dir] ;

	#
	# Keep a short history of wind directions, and then use them to create an "average" wind 
	# direction if the user wants that. Helps to counter the lack of damping in the sensor.
	#
	if (@wind_history == 0) {			# If we have no history of wind direction
		@wind_history = ($wind_dir) x 10;	# Initialise with current wind direction
		}

	my $junk = shift @wind_history;			# Remove the oldest wind direction
	push @wind_history, $wind_dir;			# Add newest wind direction	

	if ($enable_wind_avg == 1) {
		my $pi = 3.14159265358979;
		my $xvect = 0;
		my $yvect = 0;
		foreach my $junk (@wind_history) {	# Grab each wind direction in turn
			my $rad = (($junk/180) * $pi);	# Convert to radians
			$xvect += cos($rad);		# Get the vectors from the angle...
			$yvect += sin($rad);		#    and add them up
			}
		my $elements = @wind_history;
		$xvect = ($xvect / $elements);		# Now turn into average x
		$yvect = ($yvect / $elements);		# and y components
		$rad = atan2($yvect, $xvect);		# Switch the vectors back to an angle, & convert
		$wind_dir = sprintf "%.0f", (($rad / $pi) * 180) ;	# to degrees
		if ($wind_dir < 0) { $wind_dir += 360; } 	# Correct for quadrants
		}

	# Calculate Dew Point and Wind Chill using formulae from Wikipedia
	my $outdoor_dew_point = dew_point($outdoor_temperature, $outdoor_humidity);
	my $wind_chill_temp = wind_chill($outdoor_temperature, $wind_speed);

	# Total rainfall
	my $total_rain = unpack('x13S', $current_block) * 0.3;		# 0.3 for mm

	##
	## Work out rainfall rates from our various historical weather records
	##

	# Total rainfall at last (complete) weather record
	my $last_rain = unpack('x13S', $last_block) * 0.3;		# 0.3 for mm
	# Total rainfall from one hour ago weather record
	my $hour_rain = unpack('x13S', $hour_block) * 0.3;		# 0.3 for mm
	# Total rainfall from start of day weather record
	my $day_rain  = unpack('x13S', $day_block)  * 0.3;		# 0.3 for mm

	$daily_rain_rate  = ($last_rain - $day_rain) ;
	$hourly_rain_rate = ($last_rain - $hour_rain) ;

	## ############################################################################################ ##
	## At this point we have all the sensor and calculated data, in metric units (these are metric
	## weatherstations) to whatever precision we can get. Now need to display it sensibly, and
	## optionally send it to WeatherUnderground, which involves transforming to Imperial units and
	## a possible loss of accuracy.
	## ############################################################################################ ##

	##
	## Print out data in human-readable metric form, limiting to 1 decimal place as required.
	##
	if ($line >= $PAGESIZE) {
		# Print a title line every $PAGESIZE lines
		printf "%8s %3s %3s %6s %6s %6s %6s %5s %5s %3s %4s %4s %5s %6s\n",
			"Time", "%IH", "%OH", "IT\302\260C", "OT\302\260C",
			"DP\302\260C", "WC\302\260C", "Wm/s", "Gm/s", "WD",
			"1Hr", "24Hr", "Tot", "P";
		$line = 0;
		}

	printf 
		"%02d:%02d:%02d %3d %3d %5.1f %5.1f %5.1f %5.1f %5.1f %5.1f %3s %4.1f %4.1f %4.1f %5.1f\n",
		$Time[2], $Time[1], $Time[0],
		$indoor_humidity,
		$outdoor_humidity,
		$indoor_temperature,
		$outdoor_temperature,
		$outdoor_dew_point,
		$wind_chill_temp,
		$wind_speed,
		$gust_speed,
		$wind_dirs[int($wind_dir/22.5)],
		$hourly_rain_rate,
		$daily_rain_rate,
		$total_rain,
		$abs_pressure;

	# add a newline for clarity in debug mode
	print "\n" if ($debug);

	$line++;

	##
	## Upload the data to a Weather Underground PWS
	##
	if ($wuUpload == 1) { 
		if ($upload_counter <= 0){
			$upload_counter = $upload_frequency;	
			#
			# Here we will build the URL that we use to send our readings to WeatherUnderground 
			# We need to convert our data from metric to imperial:
			# from m/s -> mph
			# from degC -> degF
			# from hPa -> inches mercury
			# from mm -> inches
			#	
			my (%wuf, $url, $field);
			$wuf{dateutc} = strftime "%Y-%m-%d %H:%M:%S", gmtime;
			$wuf{winddir} = $wind_dir;
			$wuf{windspeedmph} = &mstomph($wind_speed);
			$wuf{windgustmph} = &mstomph($gust_speed);
			# Ensure gust is sensible
			if ( $wuf{windgustmph} < $wuf{windspeedmph} ) {
				$wuf{windgustmph} = $wuf{windspeedmph};
				}
			$wuf{humidity} = $outdoor_humidity;
			$wuf{tempf} = &CtoF($outdoor_temperature);
			$wuf{dewptf} = &CtoF($outdoor_dew_point);
			$wuf{dailyrainin} = &mmtoin($daily_rain_rate);
			$wuf{rainin} = &mmtoin($hourly_rain_rate);
			$wuf{baromin} = &hPatoin($abs_pressure);
			$wuf{weather} = 'NA';
			$wuf{clouds} = 'NA';

			# We very occasionally end up with bad data; lots of possible reasons for this:
			# the weather station trying to update the same memory we are trying to read, 
			# losing contact with the sensors, interference trashing wireless readings, etc
			# For me, its about 1 reading in every 50,000, but it can cause problems with 
			# WeatherUnderground reporting quality control issues on your PWS, so lets apply 
			# some sanity to what we report to them, just in case.
			if (	$wuf{windspeedmph} > 100 ||
				$wuf{windgustmph} > 120 ||
				$wuf{humidity} < 0 || $wuf{humidity} > 100 ||
				$wuf{tempf} < -60 || $wuf{tempf} > 130 ||
				$wuf{dewptf} < -60 || $wuf{dewptf} > 130 ||
				$wuf{baromin} < 12 || $wuf{baromin} > 33 ) {
				# Something isn't right, lets not report this set of data to WeatherUnderground
				foreach $field ( 'dateutc', 'baromin', 'weather', 'clouds' ) {
					$url = $url . $field . '=' . $wuf{$field} . '&';
					} 
				print "Error in: $url \n";
				}
			else {
				# Report all sensors to WeatherUnderground:
				foreach $field ( 'dateutc', 'winddir', 'windspeedmph', 'windgustmph',
						'humidity', 'tempf', 'dailyrainin', 'rainin', 'baromin', 'dewptf',
						'weather', 'clouds' ) {
					$url = $url . $field . '=' . $wuf{$field} . '&';
					} 

				# $url now holds the right data in the right format ...
				# Add the prepend and append, remove any [unwanted] return characters
				$url = $WU_URL . "?action=updateraw\&" . "ID=$ID\&" . "PASSWORD=$PASSWORD\&" . $url . "softwaretype=$SOFTWARETYPE";
				$url =~ s///g;
	
				# Comment out the fields below for testing, and uncomment the print statement
				my $ua = LWP::UserAgent->new;
				my $req = HTTP::Request->new(GET => $url);
				my $resp = $ua->simple_request($req);
				my $resp_data = $resp->content;

				# print "Return from GET: $resp \n";
				# print "Response data: $resp_data \n";
	
				# For troubleshooting:
				# print $url;
				}

			}
	
		$upload_counter--;
		}

        ### Check for end
        last if ++$loop == $loop_max;

        ### Interval
        pause;
}
